use crate::{CoreId, CpuInfo, Pid, ProcessHandle};
use core::mem::size_of;
use libc::{cpu_set_t, pid_t, sched_setaffinity};
use std::io::Error;

//#[cfg(unix)]
impl Pid {
    fn get_system_pids() -> Result<Vec<Pid>, &'static str> {
        todo!()
    }
    fn process_handle(self) -> Result<ProcessHandle<pid_t>, &'static str> {
        todo!()
    }
}

pub fn get_cpu_info() -> Result<CpuInfo, &'static str> {
    /*
    let n = num_cpus::get();
    let p = num_cpus::get_physical();
    println!("n: {}; p: {};", n, p);
    */

    let available = unsafe { libc::sysconf(libc::_SC_NPROCESSORS_CONF) };
    let online = unsafe { libc::sysconf(libc::_SC_NPROCESSORS_ONLN) };
    assert!(available > 0, "unable to detect _SC_NPROCESSORS_CONF");
    assert!(
        available <= u8::MAX as i64,
        "detected too many cpus with _SC_NPROCESSORS_CONF: {}",
        available
    );
    if online < available {
        eprintln!("TODO enable offline cpus to switch to the scheduler based optimization");
        eprintln!("mask: {}; online: {};", available, online);
    }

    Ok(CpuInfo {
        all_system_cores: (0..online).map(|i| CoreId(i as u8)).collect::<Vec<_>>(),
    })
}

pub fn set_process_affinity(name: &str, cores: &Vec<CoreId>, all_cores: &Vec<CoreId>) {
    let pids = Pid::get_system_pids();
    // println!("(DD) pids {:#?}", pids);

    let mut handles: Box<dyn Iterator<Item = ProcessHandle<pid_t>>> = Box::new(
        pids.unwrap()
            .into_iter()
            .map(Pid::process_handle)
            .filter_map(Result::ok),
    );
    if name != "*" {
        handles = Box::new(handles.filter(|p| p.name == name));
    }
    //.collect::<Vec<_>>()
    //println!("(DD) handles {:#?}", handles);
    for ph in handles {
        let cores = if ph.name.starts_with("cargo") {
            all_cores
        } else {
            cores
        };

        let set = cores_set(cores);
        let status = if 0 == unsafe { sched_setaffinity(ph.handle, size_of::<pid_t>(), &set) } {
            Err(Error::last_os_error())
        } else {
            Ok("Ok")
        };
        println!("{}: (PID: {}) <= {:?}", ph.name, ph.pid.0, status);
    }
    unimplemented!("linux set_process_affinity");
}

fn cores_set(cores: &Vec<CoreId>) -> cpu_set_t {
    todo!()
}
